@@ -2,7 +2,7 @@
 
 """package ad9xdds
 author    Benoit Dubois
-copyright FEMTO ENGINEERING, 2020
+copyright FEMTO ENGINEERING, 2018-2025
 license   GPL v3.0+
 brief     API to control AD9912 DDS development board.
 details   AD9912 development board is configurable by USB through an USB
@@ -15,20 +15,19 @@ import logging
 from array import array
 import usb.core
 from iopy.fx2 import Fx2Core
+from ad9xdds import InvalidState, bound_value
 
 
-AD_VENDOR_ID = 0x0456
-AD9912DEV_PRODUCT_ID = 0xee09
-
-VENDOR_ID = AD_VENDOR_ID
-PRODUCT_ID = AD9912DEV_PRODUCT_ID
+AD9912DEV_PRODUCT_ID_LIST = [0xee09, 0xee24]  # Dev board has plural product_id
+VENDOR_ID = 0x0456
+PRODUCT_ID = AD9912DEV_PRODUCT_ID_LIST[0]
 
 FTW_SIZE = 48               # Frequency Tuning Word register size (bit)
 PHASE_SIZE = 14             # Phase register size (bit)
 DAC_OUT_SIZE = 12           # Output DAC resolution (bit)
 
-IFMAX = 1000000000          # Input maximum frequency (Hz)
-OFMAX = 400000000           # Output maximum frequency (Hz)
+IFMAX = 1000000000.0        # Input maximum frequency (Hz)
+OFMAX = 400000000.0         # Output maximum frequency (Hz)
 AMAX = (1 << DAC_OUT_SIZE) - 1  # Output maximum amplitude (a.u.)
 
 
@@ -45,30 +44,7 @@ def split_len(seq, length):
 
 
 # =============================================================================
-def bound_value(value, vmin, vmax, val_name=""):
-    """Check that a value is included in the range [min, max], if not the value
-    is bounded to the range, ie:
-    - if value < min  ->  min = value
-    - if value > max  ->  max = value
-    :param value: Value that is checked
-    :param vmin: Minimum valid value.
-    :param vmax: Maximum valid value.
-    :returns: Bounded value.
-    """
-    if value < vmin:
-        logging.warning("Parameter %s out of range (%f). Set to: %f",
-                        val_name, value, vmin)
-        return vmin
-    elif value > vmax:
-        logging.warning("Parameter %s out of range (%f). Set to: %f",
-                        val_name, value, vmax)
-        return vmax
-    else:
-        return value
-
-
-# =============================================================================
-class Ad9912Dev(object):
+class Ad9912Dev():
     """Class representing AD9912 development card.
     Board integrates an USB port for transmission with PC and a Cypress FX2
     device is used as USB transceiver.
@@ -97,12 +73,11 @@ class Ad9912Dev(object):
 
     def __init__(self, ifreq=IFMAX):
         """The constructor.
-        :param ifreq: Current input frequency in Hz (float))
+        :param ifreq: Current input frequency in Hz (float)
         :returns: None
         """
-        super().__init__()
         self._fx2 = Fx2Core()  # USB interface
-        self._ifreq = ifreq
+        self._ifreq = float(ifreq)
         self._ofreq = None
         self._phy = None
         self._amp = None
@@ -175,12 +150,13 @@ class Ad9912Dev(object):
     def set_ifreq(self, value):
         """Set input frequency.
         :param value: Input frequency value (float)
-        :returns: None
+        :returns: ifreq value (float).
         """
-        self._ifreq = value
+        self._ifreq = float(value)
         logging.debug("Set input frequency: %r", value)
         # Update DDS device because ofreq = f(ifreq)
         self.set_ofreq(self._ofreq)
+        return value
 
     def get_ifreq(self):
         """Get input frequency.
@@ -198,6 +174,8 @@ class Ad9912Dev(object):
             else:
                 doubler = 1.0
             factor = self._pll_factor
+            if factor is None:
+                raise InvalidState("PLL multiplier factor undefined")
             sfreq = self._ifreq * doubler * factor
         else:
             sfreq = self._ifreq
@@ -210,18 +188,20 @@ class Ad9912Dev(object):
         _actual_ofreq()).
         output frequency may be (a bit) different than  The
         :param value: Output frequency value (float).
-        :returns: Actual output frequency (float)
+        :returns: Actual output frequency, -1 if process failed (float)
         """
+        value = float(value)
         if self.is_connected() is False:
             self._ofreq = value
-            return value
+            return -1
         # Compute the Frequency Tuning Word (FTW).
         ftw = int((value * (1 << FTW_SIZE)) / self.get_sysfreq())
         # Prepare list of value to send to the 6 FTW0 registers.
         ftw_val_list = split_len(format(ftw, '012x'), 2)
         # Send values to the 6 FTW0 registers of the DDS.
         for idx, ftw_val in enumerate(ftw_val_list):
-            self.set_reg(self._regName2addr['Ftw0_5']-idx, int(ftw_val, 16))
+            if self.set_reg(self._regName2addr['Ftw0_5']-idx, int(ftw_val, 16)) != 0:
+                return -1
         # Return the actual output frequency.
         self._ofreq = self._actual_ofreq(self.get_sysfreq(), ftw, FTW_SIZE)
         logging.debug("Set output frequency: %r", self._ofreq)
@@ -245,11 +225,12 @@ class Ad9912Dev(object):
         Take the queried output phase (in degree) as argument and set
         the adequat register in the DDS.
         :param value: Output phase value (float).
-        :returns: Actual output phase (float)
+        :returns: Actual output phase, -1 if process failed (float)
         """
+        value = float(value)
         if self.is_connected() is False:
             self._phy = value
-            return value
+            return -1
         # Compute dphy ("delta phase")
         dphy = int((value * (1 << PHASE_SIZE)) / 360)
         # Prepare list of value to send to the dphy registers.
@@ -257,7 +238,8 @@ class Ad9912Dev(object):
         # Send values to the dphy registers of the DDS.
         # TODO Search a better way to find reg address.
         for idx, val in enumerate(dphy_val_list):
-            self.set_reg(self._regName2addr['ADphase']-idx, int(val, 16))
+            if self.set_reg(self._regName2addr['ADphase']-idx, int(val, 16)) != 0:
+                return -1
         # Return the actual output phase
         adphy = self.actual_phy(dphy)
         logging.debug("Set phase: %r", adphy)
@@ -279,22 +261,24 @@ class Ad9912Dev(object):
         Take the input and output frequency as argument and set the adequat
         register in the DDS.
         :param value: Output amplitude value (int)
-        :returns: fsc register value if transfert is ok (int)
+        :returns: fsc register value, -1 if process failed (int)
         """
+        value = int(float(value))
         # If value is out of range, bound value and raise Warning.
-        if not(0 <= value <= AMAX):
+        if not 0 <= value <= AMAX:
             value = bound_value(value, 0, AMAX)
             logging.warning("Amplitude value out of range")
             logging.info("Amplitude set to %r", value)
         if self.is_connected() is False:
             self._amp = value
-            return value
+            return -1
         # Prepare list of value to send to the fsc registers.
         fsc_val_list = split_len(format(value, '04x'), 2)
         # Send values to the fsc registers of the DDS.
         # TODO Search a better way to find reg address.
         for idx, val in enumerate(fsc_val_list):
-            self.set_reg(self._regName2addr['DacFsMsb']-idx, int(val, 16))
+            if self.set_reg(self._regName2addr['DacFsMsb']-idx, int(val, 16)) != 0:
+                return -1
         # Return the amplitude value if transfert ok.
         logging.debug("Set output amplitude: %r", value)
         return value
@@ -343,11 +327,7 @@ class Ad9912Dev(object):
         pd_en = self.get_reg(self._regName2addr['PDEn'])
         # return the HSTL output state
         hstl = pd_en & (1 << 7)
-        if hstl == 0:
-            retval = True
-        else:
-            retval = False
-        return retval
+        return bool(hstl == 0)
 
     def set_cmos_output_state(self, state=False):
         """Set CMOS output state.
@@ -408,11 +388,7 @@ class Ad9912Dev(object):
         pd_en = self.get_reg(self._regName2addr['PDEn'])
         # return the PLL state
         doubler = pd_en & (1 << 4)
-        if doubler == 0:
-            retval = True
-        else:
-            retval = False
-        return retval
+        return bool(doubler == 0)
 
     def set_pll_doubler_state(self, flag=False):
         """Set SysClk PLL doubler state.
@@ -432,7 +408,7 @@ class Ad9912Dev(object):
                 msg = reg & mask
             self.set_reg(self._regName2addr['PllParam'], msg)
         self._pll_doubler_state = flag
-        logging.debug("Set PLL doubler state to: " + str(flag))
+        logging.debug("Set PLL doubler state to: %r", flag)
 
     def get_pll_doubler_state(self):
         """Get PLL doubler state.
@@ -442,11 +418,7 @@ class Ad9912Dev(object):
         pll_param = self.get_reg(self._regName2addr['PllParam'])
         # return the PLL doubler state
         doubler = pll_param & (1 << 3)
-        if doubler == 1 << 3:
-            retval = True
-        else:
-            retval = False
-        return retval
+        return bool(doubler == 1 << 3)
 
     def set_pll_multiplier_factor(self, factor):
         """Set SysClk PLL multiplier factor.
@@ -463,7 +435,7 @@ class Ad9912Dev(object):
             msg = factor / 2  # To take account of prescaler
             self.set_reg(self._regName2addr['Ndivider'], msg)
         self._pll_factor = factor
-        logging.debug("Set PLL multiplier factor to: " + str(factor))
+        logging.debug("Set PLL multiplier factor to: %r", factor)
 
     def get_pll_multiplier_factor(self):
         """Get SysClk PLL multiplier factor.
@@ -489,7 +461,7 @@ class Ad9912Dev(object):
         reg = reg & ~((1 << 1) + (1 << 0))
         msg = reg + value
         self.set_reg(self._regName2addr['PllParam'], msg)
-        logging.debug("Set charge pump current to: " + str(value))
+        logging.debug("Set charge pump current to: %r", value)
 
     def get_cp_current(self):
         """Get charge pump current configuration value.
@@ -533,7 +505,7 @@ class Ad9912Dev(object):
             mask = 1 << 7
             msg = reg | mask
         self.set_reg(self._regName2addr['PllParam'], msg)
-        logging.debug("Set VCO range to: " + str(value))
+        logging.debug("Set VCO range to: %r", value)
 
     def get_vco_range(self):
         """Get VCO range configuration value.
@@ -584,7 +556,7 @@ class Ad9912Dev(object):
             msg = reg & mask
         self.set_reg(self._regName2addr['PDEn'], msg)
         #
-        logging.debug("Set HSTL doubler state to: " + str(flag))
+        logging.debug("Set HSTL doubler state to: %r", flag)
 
     def get_hstl_doubler_state(self):
         """Get HSTL doubler state.
@@ -594,11 +566,7 @@ class Ad9912Dev(object):
         pll_param = self.get_reg(self._regName2addr['HstlDrv'])
         # return the PLL doubler state
         doubler = pll_param & (1 << 1) & (1 << 0)
-        if doubler == 2:
-            retval = True
-        else:
-            retval = False
-        return retval
+        return bool(doubler == 2)
 
     def set_led(self, flag=False):
         """This gives the user control over the USB Status LED on the board.
@@ -616,13 +584,13 @@ class Ad9912Dev(object):
         try:
             self._fx2.write_ctrl(msg)
         except usb.core.USBError as ex:
-            logging.error("Could set LED blink: " + str(ex))
+            logging.error("Could set LED blink: %r", ex)
 
     def set_reg(self, address, value):
         """Set register value @ address.
         :param address: Address of the register to write (int).
         :param value: Value to write (int).
-        :returns: None
+        :returns: 0 if ok else -1 (int)
         """
         value = int(value)
         # Build a list of binary value corresponding to the instruction
@@ -635,17 +603,29 @@ class Ad9912Dev(object):
         for idx, val in enumerate(bword):
             if val == '1':
                 msg[len(msg)-1-idx] = 1
-        # Write real data
-        self._fx2.write(msg)
-        # End write cycle (Set value to portA's pins of FX2 device)
-        try:
-            end = array('B', (0x01, 0x0B))
-            self._fx2._ep_out.write(end)
-            end = array('B', (0x01, 0x03))
-            self._fx2._ep_out.write(end)
-        except usb.core.USBError as ex:
-            logging.error("Could not set " + str(value) + " @address "
-                          + str(address) + ", error: " + str(ex))
+        write_flag = False
+        while write_flag is False:
+            # Write real data
+            try:
+                self._fx2.write(msg)
+            except usb.core.USBError as ex:
+                logging.error("Write msg %r failed: %r", msg, ex)
+                return -1
+            # End write cycle (Set value to portA's pins of FX2 device)
+            try:
+                end = array('B', (0x01, 0x0B))
+                self._fx2._ep_out.write(end)
+                end = array('B', (0x01, 0x03))
+                self._fx2._ep_out.write(end)
+            except usb.core.USBError as ex:
+                logging.error("Set %r @address %r failed: %r",
+                              value, address, ex)
+                return -1
+            reg_value = self.get_reg(address)
+            if reg_value == value:
+                write_flag = True
+        return 0
+
 
     def get_reg(self, address):
         """Get register value @ address.
@@ -657,7 +637,8 @@ class Ad9912Dev(object):
         try:
             self._fx2._ep_out.write(init)
         except usb.core.USBError as ex:
-            logging.error("Could not init get register cycle: %s", str(ex))
+            logging.error("Get register cycle failed: %r", ex)
+            return -1
         # Query register reading operation
         bword = self._build_instruct(1, 0, address)
         bword = (bin(bword)).replace('0b', '')
@@ -667,15 +648,24 @@ class Ad9912Dev(object):
         for idx, val in enumerate(bword):
             if val == '1':
                 msg[len(msg)-1-idx] = 1
-        self._fx2.write(msg)
+        try:
+            self._fx2.write(msg)
+        except usb.core.USBError as ex:
+            logging.error("Write msg %r failed: %r", msg, ex)
+            return -1
         # Read result.
-        raw_data = self._fx2.read()
+        try:
+            raw_data = self._fx2.read()
+        except usb.core.USBError as ex:
+            logging.error("Read raw_data failed: %r", ex)
+            return -1
         # Disable single byte read mode (readback mode).
         end = array('B', (0x04, 0x00))
         try:
             self._fx2._ep_out.write(end)
         except usb.core.USBError as ex:
-            logging.error("Could not finish get register cycle: %s", str(ex))
+            logging.error("Get register cycle failed: %r", ex)
+            return -1
         # Build the register value from the raw data.
         result = 0
         for idx, val in enumerate(raw_data):
